/* Agent Tcl
   Bob Gray
   1 September 1995

   genQueue.cc

   This file implements a FIFO or queue.

   Copyright (c) 1995-1997, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "genQueue.h"

/*
 * initial and minimum queue size
 */

const int DEFAULT_QUEUE_SIZE = 8;
const int MINIMUM_QUEUE_SIZE = 8;

  /* constructor for class QUEUE */

QUEUE::QUEUE (void)
{
  elements = new void * [DEFAULT_QUEUE_SIZE];
   
    /* number of elements and number of slots */

  count = 0;
  slots = DEFAULT_QUEUE_SIZE;

    /* expand and shrink count */

  expand_count = slots;

  if (slots <= MINIMUM_QUEUE_SIZE) {
    shrink_count = -1;
  } else {
    shrink_count = slots / 4;
  }

    /* head and tail indices */

  tail = 0;
  head = 0;
}

  /* destructor for class QUEUE */

QUEUE::~QUEUE ()
{
  delete elements;
}

/* QUEUE::resize_queue

   Purpose: Expand or shrink the queue

     Input: which = e_QUEUE_SHRINK or e_QUEUE_EXPAND

    Output: The procedure doubles or halves the size of the queue.
*/

void QUEUE::resize_queue (QueueSizing which)
{
  int new_slots;
  void **new_elements;

  if (which == e_QUEUE_EXPAND) {
    new_slots = slots << 1;
  } else {
    new_slots = slots >> 1;
  }

  new_elements = new void * [new_slots];

    /* copy over the queue elements */

  for (int i = 0; i < count; i++) {
    new_elements[i] = elements[head];
    head = (head + 1) % slots;
  }

  delete elements;

    /* reinitialize all the data fields */

  elements     = new_elements;
  slots        = new_slots;
  head         = 0;
  tail         = count;
  expand_count = new_slots;

  if (slots <= MINIMUM_QUEUE_SIZE) {
    shrink_count = -1;
  } else {
    shrink_count = slots / 4;
  }
}

/* QUEUE::enqueue

   Purpose: Add an element to the the tail of the queue

     Input: item = item to add to the tail of the queue

    Output: The procedure adds the item to the queue.
*/

void QUEUE::enqueue (void *item)
{
    /* assertions on the parameters */

  assert (item != NULL);
  
    /* add the item to the queue */

  elements[tail] = item;
  tail           = (tail + 1) % slots;
  count         += 1;

    /* expand the queue if necessary */

  if (count >= expand_count) {
    resize_queue (e_QUEUE_EXPAND);
  }
}

/* QUEUE::push

   Purpose: Add an element to the the head of the queue

     Input: item = item to add to the head of the queue

    Output: The procedure adds the item to the queue.
*/

void QUEUE::push (void *item)
{
    /* assertions on the parameters */

  assert (item != NULL);
  
    /* add the item to the queue */

  head           = (head - 1) % slots;
  elements[head] = item;
  count         += 1;

    /* expand the queue if necessary */

  if (count >= expand_count) {
    resize_queue (e_QUEUE_EXPAND);
  }
}

/* QUEUE::dequeue

   Purpose: Get the item at the head of the queue

     Input: None

    Output: The procedure returns NULL if the queue is empty.  Otherwise the
            procedure returns a pointer to the item at the head of the queue
            and removes that item from the queue.
*/

void *QUEUE::dequeue (void)
{
  register void *item;

    /* check if the queue is empty */

  if (head == tail) {
    return NULL;
  }

    /* get the head of the queue */

  item   = elements[head];
  head   = (head + 1) % slots;
  count -= 1;  

    /* shrink the queue if necessary */

  if (count <= shrink_count) {
    resize_queue (e_QUEUE_SHRINK); 
  }

    /* return the head of the queue */

  return (item);
}

/* QUEUE::peek

   Purpose: Peek at an item in the queue

     Input: i = position of the item within the queue

    Output: The procedure returns NULL if the queue has i or fewer items.
            Otherwise the procedure returns a pointer to the i TH element in
            the queue.
*/

void *QUEUE::peek (int i) const
{
  register int temp;

  if (i >= count) {
    return NULL;
  } else {
    temp = (head + i) % slots; 
    return (elements[temp]);
  }
}
